/* 
 * SI - Sound Interface
 * Copyright (C) 2002 Matej Knopp
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "gem-list.h"
#include "si-config.h"

/* 
 * Simple implementation (without memory pools)
 */

#define gem_malloc0 si_malloc0
#define gem_free    si_free

static inline GemList *
gem_list_do_alloc (void)
{
	GemList *list;
	list = gem_malloc0 (sizeof (GemList));
	return list;
}

GemList *
gem_list_alloc (void)
{
	return gem_list_do_alloc ();
}


void
gem_list_free_node (GemList * list)
{
	gem_free (list);
}

void
gem_list_free (GemList * list)
{
	GemList *cur;
	
	while (list)
	{
		cur = list;
		list = list->next;
		gem_free (cur);		
	}
}

GemList *
gem_list_prepend (GemList * list, void * data)
{
	GemList *new_list;
	
	new_list = gem_list_do_alloc ();
	new_list->data = data;
	
	if (list == NULL)
	{
		return new_list;
	}

	while (list->prev)
	{
		list = list->prev;
	}
	
	list->prev = new_list;
	new_list->next = list;
	
	return new_list;
}

GemList *
gem_list_append (GemList * list, void * data)
{
	GemList *new_list;
	GemList *last;
	
	new_list = gem_list_do_alloc ();
	new_list->data = data;
	
	if (list == NULL)
	{
		return new_list;
	}

	last = list;

	while (last->next)
	{
		last = last->next;
	}
	
	last->next = new_list;
	new_list->prev = last;
	
	return list;
}

GemList *
gem_list_insert (GemList * list, int pos, void * data)
{
	GemList *new_list;
	GemList *cur;
	GemList *tmp;
	
	if (pos == -1)
		return gem_list_append (list, data);
	else if (pos == 0)
		return gem_list_prepend (list, data);
		
	new_list = gem_list_do_alloc ();
	new_list->data = data;
	if (list == NULL)
	{
		return new_list;
	}
	
	cur = list;	
	
	while ((pos - 1) > 0 && cur->next)
	{
		cur = cur->next;
		-- pos;
	}
	
	tmp = cur->next;
	cur->next = new_list;
	new_list->prev = cur;
	new_list->next = tmp;
	if (tmp)
		tmp->prev = new_list;
	
	return list;
}

GemList *
gem_list_insert_after (GemList * list, GemList * node, void * data)
{
	GemList *new_list;
	GemList *tmp;

	if (list && !node)
	{
		return gem_list_append (list, data);
	}
	
	new_list = gem_list_do_alloc ();
	new_list->data = data;

	if (!list)
	{
		si_return_val_if_fail (node == NULL, new_list);
		return new_list;
	}

	tmp = node->next;
	node->next = new_list;
	new_list->prev = node;
	new_list->next = tmp;
	if (tmp)
		tmp->prev = new_list;
	return list;
}

GemList *
gem_list_insert_before (GemList * list, GemList * node, void * data)
{
	GemList *new_list;
	GemList *tmp;

	if (list && !node)
	{
		return gem_list_prepend (list, data);
	}
	
	new_list = gem_list_do_alloc ();
	new_list->data = data;

	if (!list)
	{
		si_return_val_if_fail (node == NULL, new_list);
		return new_list;
	}

	tmp = node->prev;
	node->prev = new_list;
	new_list->next = node;
	new_list->prev = tmp;
	if (tmp)
		tmp->next = new_list;
		
	/* Return the first item in list */
	return node != list ? list : list->prev;
}

GemList *
gem_list_insert_sorted (GemList * list, void * data, GemCompareFunc comp)
{
	GemList *new_list;
	GemList *cur;
	int res;
	
	new_list = gem_list_do_alloc ();
	new_list->data = data;
	
	if (list == NULL)
		return new_list;
	
	cur = list;
	res = (* comp) (data, cur->data);
	while (cur->next && (res > 0))
	{
		cur = cur->next;
		res = (* comp) (data, cur->data);
	}
	
	if (res > 0) /* reached last - append */
	{
		cur->next = new_list;
		new_list->prev = cur;
		return list;
	}
	
	if (cur->prev)
	{
		cur->prev->next = new_list;
		new_list->prev = cur->prev;
	}
	cur->prev = new_list;
	new_list->next = cur;
	
	return list == cur ? new_list : list;
}

GemList *
gem_list_node_at (GemList * list, int pos)
{
	if (list && pos >= 0)
	{
		while (pos-- && list != NULL)
		{
			list = list->next;
		}
		return list;
	}
	return NULL;
}

void *
gem_list_data_at (GemList * list, int pos)
{
	if (list && pos >= 0)
	{
		while (pos-- && list != NULL)
		{
			list = list->next;
		}
		return list ? list->data : NULL;
	}
	return NULL;
}

GemList *
gem_list_remove_node (GemList * list, GemList * node)
{
	if (node)
	{
		if (node->prev)
			node->prev->next = node->next;
		if (node->next)
			node->next->prev = node->prev;
		if (node == list)
			list = list->next;

		node->prev = node->next = NULL;
	}
	return list;
}

GemList *
gem_list_delete_node (GemList * list, GemList * node)
{
	if (node)
	{
		if (node->prev)
			node->prev->next = node->next;
		if (node->next)
			node->next->prev = node->prev;
		if (node == list)
			list = list->next;

		gem_list_free_node (node);
	}
	return list;
}

GemList *
gem_list_remove (GemList * list, const void * data)
{
	GemList *node = list;
	
	while (node)
	{
		if (node->data != data)
		{
			node = node->next;
		}
		else
		{
			if (node->prev)
				node->prev->next = node->next;
			if (node->next)
				node->next->prev = node->prev;
			if (node == list)
				list = list->next;

			gem_list_free_node (node);
			break;
		}
	}
	
	return list;
}

GemList *
gem_list_remove_all (GemList * list, const void * data)
{
	GemList *node = list;
	GemList *tmp;
	
	while (node)
	{
		if (node->data != data)
		{
			node = node->next;
		}
		else
		{
			tmp = node;
			if (node->prev)
				node->prev->next = node->next;
			if (node->next)
				node->next->prev = node->prev;
			if (node == list)
				list = list->next;
			node = node->next;
			gem_list_free_node (tmp);
		}
	}
	
	return list;
}

GemList *
gem_list_concat (GemList * list1, GemList * list2)
{
	GemList *first = list1;

	if (list1 == NULL)
		return list2;
	if (list2 == NULL)
		return list1;

	while (list1->next)
		list1 = list1->next;

	list1->next = list2;
	list2->prev = list1;
	
	return first;
}

GemList *
gem_list_copy (GemList * list)
{
	GemList *new_list;
	GemList *last = NULL;
	GemList *first = NULL;
	
	while (list)
	{
		new_list = gem_list_do_alloc ();
		new_list->data = list->data;
		new_list->prev = last;
		if (last)
			last->next = new_list;
		last = new_list;
		
		if (!first)
			first = new_list;
		
		list = list->next;
	}
	return first;
}

GemList *
gem_list_reverse (GemList * list)
{
	GemList *last = NULL;
	
	while (list)
	{
		last = list;
		list = list->next;
		last->next = last->prev;
		last->prev = list;
	}

	return last;
}

GemList *
gem_list_first (GemList * list)
{
	while (list && list->prev)
		list = list->prev;
	return list;
}

GemList *
gem_list_last (GemList * list)
{
	while (list && list->next)
		list = list->next;
	return list;
}

unsigned
gem_list_length (GemList * list)
{
	unsigned len = 0;
	while (list)
	{
		++ len;
		list = list->next;
	}
	return len;
}

void
gem_list_foreach (GemList * list, void (*func) (void * data))
{
	while (list)
	{
		(* func) (list->data);
		list = list->next;
	}
}

GemList *
gem_list_find (GemList * list, const void * data)
{
	while (list)
	{
		if (list->data == data)
			break;
		list = list->next;
	}
	return list;
}

GemList *
gem_list_find_custom (GemList * list, const void * data, GemCompareFunc comp)
{
	while (list)
	{
		if ((* comp) (list->data, data) == 0)
			break;
		list = list->next;
	}
	return list;
}

/*
 * The sorting algorithm is taken from GLib <http://www.gtk.org>
 */

static GemList *
gem_list_sort_merge (GemList * l1, GemList * l2, GemCompareFunc comp)
{
	GemList list, *l, *lprev;
	int cmp;

	l = &list; 
	lprev = NULL;

	while (l1 && l2)
	{
		cmp = (* comp) (l1->data, l2->data);
		
		if (cmp <= 0)
		{
			l->next = l1;
			l = l->next;
			l->prev = lprev; 
			lprev = l;
			l1 = l1->next;
		} 
		else 
		{
			l->next = l2;
			l = l->next;
			l->prev = lprev; 
			lprev = l;
			l2 = l2->next;
		}
	}
	l->next = l1 ? l1 : l2;
	l->next->prev = l;

  	return list.next;
}

GemList * 
gem_list_sort (GemList * list, GemCompareFunc comp)
{
	GemList *l1, *l2;
  
	if (!list) 
    	return NULL;
	if (!list->next) 
    	return list;
  
	l1 = list; 
	l2 = list->next;

	while ((l2 = l2->next) != NULL)
	{
		if ((l2 = l2->next) == NULL) 
			break;
		l1 = l1->next;
	}
	l2 = l1->next; 
	l1->next = NULL; 

	return gem_list_sort_merge (gem_list_sort (list, comp),
	                            gem_list_sort (l2, comp), comp);
}


