/* 
 * SI - Sound Interface
 * Copyright (c) 2002 Matej Knopp
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */
 
#include "si.h"
#include "si-internal.h"

#include <stdlib.h>
#include <string.h>

static const char * extensions [] =
{
	"raw",
	NULL
};

SiLoaderInfo _si_raw_info =
{
	"raw", /* id */
	extensions, /* supported extensions (null terminated) */
	"RAW file loader", /* short description */
	"Matej Knopp <knopp@users.sf.net>", /* author (+e-mail) */
	"http://" /* url */
};

static int 
raw_init (void)
{
#ifdef DEBUG_LOADER
	fprintf (stderr, "raw_init ()\n");
#endif
	return 1;
}

static void
raw_quit (void)
{
#ifdef DEBUG_LOADER
	fprintf (stderr, "raw_quit ()\n");
#endif
}

static int
raw_open (SiSoundData * sound_data, const char * extension,
          const SiLoaderArgument * arguments)
{
	const SiLoaderArgument *a;
	SiAudioInfo fmt = { 0, 0, 0 };

#ifdef DEBUG_LOADER
	fprintf (stderr, "raw_open ()\n");	
#endif

	/* accept only files with explicitely given extension */
	if (si_strcasecmp (extension, "raw") != 0)
		return 0;

	a = arguments;
	/* the format of raw files must be specified within arguments */
	while (a && a->key)
	{
		if (si_strcasecmp (a->key, "format") == 0)
		{
#define _FMT(_format) if (si_strcasecmp (a->value, #_format) == 0) fmt.format = SI_AUDIO_##_format;
			_FMT (U8);     _FMT (S8); 
			_FMT (U16LSB); _FMT (S16LSB);
			_FMT (U16MSB); _FMT (S16MSB);
			_FMT (U16SYS); _FMT (S16SYS);
#undef _FMT
		}
		else if (si_strcasecmp (a->key, "rate") == 0)
		{
			fmt.rate = atoi (a->value);
		}
		else if (si_strcasecmp (a->key, "channels") == 0)
		{
			fmt.channels = atoi (a->value);
		}
		a++;
	}
	
	if (!fmt.format || !fmt.rate || !fmt.channels)
		return 0;

	sound_data->internal->actual = fmt;
	
	return 1;
}

static void
raw_close (SiSoundData * sound_data)
{
#ifdef DEBUG_LOADER
	fprintf (stderr, "raw_close ()\n");
#endif
	/* we don't need to free anything since we had allocated nothing */	
}

static Uint32
raw_decode (SiSoundData * sound_data)
{
#ifdef DEBUG_LOADER
	/* fprintf (stderr, "raw_decode ()\n"); */
#endif	
	if (sound_data->flags & SI_SOUND_DATA_STREAM)
	{
		Uint32 num_read =
			si_read (sound_data, sound_data->internal->buffer, 1,
			         sound_data->internal->buffer_size);

		if (num_read < sound_data->internal->buffer_size)
			sound_data->flags |= SI_SOUND_DATA_EOF;
			
		return num_read;
	}
	else
	{
		int file_size;
		int read;
		
		file_size = si_seek (sound_data, 0, SEEK_END);
			
		sound_data->internal->buffer_size = file_size;
		sound_data->internal->buffer = si_malloc (file_size);
			
		if (!sound_data->internal->buffer)
		{
			sound_data->flags |= SI_SOUND_DATA_ERROR;
			si_set_error (SI_ERROR_MEM);
			return 0;
		}
		
		si_seek (sound_data, 0, SEEK_SET);

		read = si_read (sound_data, sound_data->internal->buffer, 1, 
		                file_size);
	
		sound_data->flags |= SI_SOUND_DATA_EOF;

		return read;
	}
}

static int
raw_rewind (SiSoundData * sound_data)
{
#ifdef DEBUG_LOADER
	printf ("raw_rewind ()\n");
#endif	
	si_seek (sound_data, 0, SEEK_SET);
	return 1;
}

SiLoaderInternal _si_raw_internal =
{
	raw_init,
	raw_quit,
	raw_open,
	raw_close,
	raw_decode,
	raw_rewind
};

