/* 
 * SI - Sound Interface
 * Copyright (c) 2002 Matej Knopp
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */
 
/*
 * RIFF WAVE LOADER
 * ----------------
 * Supports MS ADPCM and IMA ADPCM compression.
 *
 * Some (compression related) code is taken from the SDL_sound library
 * and the SDL library.
 */

/*
 * !FIXME! - The SDL_Sound and SDL Copyright here!
 */

#include "si.h"
#include "si-internal.h"

#include <stdlib.h>
#include <string.h>

//#define DEBUG_LOADER

static const char * extensions [] =
{
	"wav",
	NULL
};

SiLoaderInfo _si_wav_info =
{
	"wav", /* id */
	extensions, /* supported extensions (null terminated) */
	"WAVE file loader", /* short description */
	"Matej Knopp <knopp@users.sf.net>", /* author (+e-mail) */
	"http://" /* url */
};

static int 
wav_init (void)
{
#ifdef DEBUG_LOADER
	fprintf (stderr, "wav_init ()\n");
#endif
	return 1;
}

static void
wav_quit (void)
{
#ifdef DEBUG_LOADER
	fprintf (stderr, "wav_quit ()\n");
#endif
}

#define BAIL(arg, error, data, ret) if (!(arg)) { si_set_error (error); data->flags |= SI_SOUND_DATA_ERROR; return ret; }
#define BAIL_IO(arg, data, ret) BAIL (arg, SI_ERROR_IO, data, ret);
#define BAIL_FMT(arg, data, ret) BAIL (arg, "Incorrect wav format.", data, ret);
#define BAIL_MEM(arg, data, ret) BAIL (arg, SI_ERROR_MEM, data, ret);

static int 
wav_find_chunk (SiSoundData * data, const char _ID[4], int from_begin)
{
	char ID [4];
	Uint32 len;

	if (from_begin)
		si_seek (data, 12, SEEK_SET); /* move behind the header */

	BAIL_FMT (si_read (data, ID, 4, 1), data, 0);

	while (memcmp (_ID,ID,4) != 0) 
	{
		BAIL_FMT (si_read_le32_ok (data, &len), data, 0)
		si_seek (data, len, SEEK_CUR);
		BAIL_FMT (si_read (data, ID, 4, 1), data, 0);
	}

	return 1;
}

typedef struct _MsAdpcmCoefSet      MsAdpcmCoefSet;
typedef struct _MsAdpcmBlockHeader  MsAdpcmBlockHeader;
typedef struct _ImaAdpcmBlockHeader ImaAdpcmBlockHeader;

struct _MsAdpcmCoefSet
{
	Sint16 coef1;
	Sint16 coef2;
};

struct _MsAdpcmBlockHeader 
{
	Uint8  predictor;
	Uint16 delta;
	Sint16 samp1;
	Sint16 samp2;
};

struct _ImaAdpcmBlockHeader
{
	Sint32 sample;
	Sint8  index;
	Sint8  reserved;
};

enum  /* Values of format.tag */
{
	WAVE_FMT_NORMAL     = 0x01,
	WAVE_FMT_MS_ADPCM   = 0x02,
	WAVE_FMT_IMA_ADPCM  = 0x11,
};

struct _SiLoaderPrivate
{
	struct
	{
		Sint16 tag;
		Uint16 channels;
		Uint32 samples_per_sec;
		Uint32 avg_bytes_per_sec;
		Uint16 block_align;
		Uint16 bits_per_sample;
	} format;

	Uint32 bytes_left;
	Uint32 decompressed_size; /* this is stored in fact chunk */
	Uint16 sample_frame_size;
	
	union /* additional information (depends on format.tag) */
	{
		struct
		{
			Uint16               size;
			Uint16               samples_per_block;
			Uint16               num_coef;
			MsAdpcmCoefSet     * coef;
			MsAdpcmBlockHeader * block_headers;
			Uint32               samples_left_in_block;
			Uint8              * block;
			Uint32               block_pos;
			int                  nibble_state;
			Sint8                nibble;
		} ms_adpcm;
		struct
		{
			Uint16                size;
			Uint16                samples_per_block;
			ImaAdpcmBlockHeader * block_headers;
			Uint32                samples_left_in_block;
			Uint8               * decoded;
			Uint16                decoded_size;
			Uint8                 last_copied;
		} ima_adpcm;
	};
	
	/* free the allocated additional information */
	void   (* free)   (SiSoundData * sound_data);
	Uint32 (* decode) (SiSoundData * sound_data);
	int    (* rewind) (SiSoundData * sound_data);
};

static Uint32 wav_decode_normal	   (SiSoundData * sound_data);
static int    wav_rewind_normal    (SiSoundData * sound_data);

static void   wav_free_ms_adpcm	   (SiSoundData * sound_data);
static Uint32 wav_decode_ms_adpcm  (SiSoundData * sound_data);
static int    wav_rewind_ms_adpcm  (SiSoundData * sound_data);

static void   wav_free_ima_adpcm   (SiSoundData * sound_data);
static Uint32 wav_decode_ima_adpcm (SiSoundData * sound_data);
static int    wav_rewind_ima_adpcm (SiSoundData * sound_data);

static int
wav_read_fmt_generic (SiSoundData * data)
{
	SiLoaderPrivate *priv = data->internal->loader_private;
	
	Uint32 ch_len; /* chunk length */
	
	BAIL_IO (si_read_le32_ok (data, &ch_len), data, 0);
	BAIL_IO (si_read_le16_ok (data, &priv->format.tag), data, 0);
	BAIL_IO (si_read_le16_ok (data, &priv->format.channels), data, 0);
	BAIL_IO (si_read_le32_ok (data, &priv->format.samples_per_sec), data, 0);
	BAIL_IO (si_read_le32_ok (data, &priv->format.avg_bytes_per_sec), data, 0);
	BAIL_IO (si_read_le16_ok (data, &priv->format.block_align), data, 0);
	BAIL_IO (si_read_le16_ok (data, &priv->format.bits_per_sample), data, 0);

	data->internal->actual.channels = (Uint8) priv->format.channels;
	data->internal->actual.rate = priv->format.samples_per_sec;

	if (priv->format.bits_per_sample == 8)
		data->internal->actual.format = SI_AUDIO_U8;
	else if (priv->format.bits_per_sample == 16)
		data->internal->actual.format = SI_AUDIO_S16LSB;
	else if (priv->format.bits_per_sample == 4) /* adpcm */
		data->internal->actual.format = SI_AUDIO_S16SYS;
	else 
	{
		si_set_error ("Wave must be 8, 16 or 4 bit.");
		data->flags |= SI_SOUND_DATA_ERROR;
		return 0;
	}

	return 1;
};

static int
wav_read_fmt_normal (SiSoundData * data)
{
	/* not much to be done */
	SiLoaderPrivate *priv = data->internal->loader_private;
	priv->free   = NULL;
	priv->decode = wav_decode_normal;
	priv->rewind = wav_rewind_normal;

	return 1;
}

static int
wav_read_fmt_ms_adpcm (SiSoundData * data)
{
	int i;

	SiLoaderPrivate *priv = data->internal->loader_private;	

	priv->free   = wav_free_ms_adpcm;
	priv->decode = wav_decode_ms_adpcm;
	priv->rewind = wav_rewind_ms_adpcm;

	BAIL_IO (si_read_le16_ok (data, &priv->ms_adpcm.size), data, 0);
	BAIL_IO (si_read_le16_ok (data, &priv->ms_adpcm.samples_per_block), data, 0);
	BAIL_IO (si_read_le16_ok (data, &priv->ms_adpcm.num_coef), data, 0);

	/* read coefficients */
	priv->ms_adpcm.coef = si_new0 (MsAdpcmCoefSet, priv->ms_adpcm.num_coef);
	BAIL_MEM (priv->ms_adpcm.coef != NULL, data, 0);
	
	for (i = 0; i < priv->ms_adpcm.num_coef; ++i)
	{
		BAIL_IO (si_read_le16_ok (data, &priv->ms_adpcm.coef [i].coef1), data, 0);
		BAIL_IO (si_read_le16_ok (data, &priv->ms_adpcm.coef [i].coef2), data, 0);
	}

	priv->ms_adpcm.block_headers = si_new0 (MsAdpcmBlockHeader,
						priv->format.channels);
	if (priv->ms_adpcm.block_headers == NULL)
	{
		si_set_error (SI_ERROR_MEM);
		si_free (priv->ms_adpcm.coef);
		return 0;
	}

	priv->ms_adpcm.block = si_malloc (priv->ms_adpcm.samples_per_block);

	if (priv->ms_adpcm.block == NULL)
	{
		si_set_error (SI_ERROR_MEM);
		si_free (priv->ms_adpcm.coef);
		si_free (priv->ms_adpcm.block_headers);
		return 0;
	}
	
	return 1;
}

static int
wav_read_fmt_ima_adpcm (SiSoundData * data)
{
	SiLoaderPrivate *priv = data->internal->loader_private;	

	priv->free   = wav_free_ima_adpcm;
	priv->decode = wav_decode_ima_adpcm;
	priv->rewind = wav_rewind_ima_adpcm;

	BAIL_IO (si_read_le16_ok (data, &priv->ima_adpcm.size), data, 0);
	BAIL_IO (si_read_le16_ok (data, &priv->ima_adpcm.samples_per_block), data, 0);

	priv->ima_adpcm.block_headers = 
		si_new0 (ImaAdpcmBlockHeader, priv->format.channels);
	BAIL_MEM (priv->ima_adpcm.block_headers != NULL, data, 0);

	priv->ima_adpcm.decoded_size = priv->format.channels * 16;
	priv->ima_adpcm.decoded = si_malloc0 (priv->format.channels * 16);
	if (priv->ima_adpcm.decoded == NULL)
	{
		si_set_error (SI_ERROR_MEM);
		si_free (priv->ima_adpcm.block_headers);
		return 0;
	}	
	
	return 1;
}


static int
wav_read_fmt (SiSoundData * data)
{
	SiLoaderPrivate *priv = data->internal->loader_private;

	if (!wav_read_fmt_generic (data))
		return 0;
	
	switch (priv->format.tag)
	{
	case WAVE_FMT_NORMAL:
		{
			if (!wav_read_fmt_normal (data))
				return 0;
			return 1;
		}
	case WAVE_FMT_MS_ADPCM:
		{
			if (!wav_read_fmt_ms_adpcm (data))
				return 0;
			return 1;
		}
	case WAVE_FMT_IMA_ADPCM:
		{
			if (!wav_read_fmt_ima_adpcm (data))
				return 0;
			return 1;
		}
	default:
		si_set_error ("Unsupported wave type.");
		data->flags |= SI_SOUND_DATA_ERROR;
		return 0;
	}
}

static int
wav_open (SiSoundData * sound_data, const char * extension,
          const SiLoaderArgument * arguments)
{
	char header [12];
	SiLoaderPrivate *priv;
	
#ifdef DEBUG_LOADER
	fprintf (stderr, "wav_open ()\n");	
#endif
	if (si_read (sound_data, header, 4, 1) == 0)
		return 0;
	if (memcmp (header, "RIFF", 4) != 0)
		return 0;
	/* we have RIFF wave file */
	
	/* finish the header */
	BAIL_IO  (si_read (sound_data, header + 4, 8, 1), sound_data, 0);
	BAIL_FMT (memcmp (header + 8, "WAVE", 4) == 0, sound_data, 0);

	priv = si_new0 (SiLoaderPrivate, 1);
	BAIL_MEM (priv != 0, sound_data, 0);

	sound_data->internal->loader_private = priv;

	if (!wav_find_chunk (sound_data, "fmt ", 0) ||
	    !wav_read_fmt (sound_data))
	{
		if (priv->free)
			priv->free (sound_data);
		si_free (priv);
		return 0;
	}

	if (priv->format.tag != WAVE_FMT_NORMAL)
	{
		Uint32 chunk_size;
		/* must have fact chunk */
		if (!wav_find_chunk (sound_data, "fact", 1) ||
		    !si_read_le32_ok (sound_data, &chunk_size) ||
		    !si_read_le32_ok (sound_data, &priv->decompressed_size))
		{
			if (priv->free)
				priv->free (sound_data);
			si_free (priv);
			si_set_error (SI_ERROR_IO);
			return 0;
		}
		/* seek to the end of the chunk */
		si_seek (sound_data, chunk_size - 4, SEEK_CUR);
	}
	if (!wav_find_chunk (sound_data, "data",0) ||
            !si_read_le32_ok (sound_data, &priv->bytes_left))
	{
		if (priv->free)
			priv->free (sound_data);
		si_free (priv);
		si_set_error (SI_ERROR_IO);
		return 0;
	}
	priv->sample_frame_size = 
		(((sound_data->internal->actual.format & 0xFF) / 8) *
                 sound_data->internal->actual.channels );
	return 1;
}

static void
wav_free_ms_adpcm (SiSoundData * sound_data)
{
	SiLoaderPrivate *priv = sound_data->internal->loader_private;
	if (priv->ms_adpcm.coef)
		si_free (priv->ms_adpcm.coef);
	if (priv->ms_adpcm.block_headers)
		si_free (priv->ms_adpcm.block_headers);
	if (priv->ms_adpcm.block)
		si_free (priv->ms_adpcm.block);
}

static void
wav_free_ima_adpcm (SiSoundData * sound_data)
{
	SiLoaderPrivate *priv = sound_data->internal->loader_private;
	if (priv->ima_adpcm.block_headers)
		si_free (priv->ima_adpcm.block_headers);
	if (priv->ima_adpcm.decoded)
		si_free (priv->ima_adpcm.decoded);
}

static void
wav_close (SiSoundData * sound_data)
{
	SiLoaderPrivate *priv = sound_data->internal->loader_private;
#ifdef DEBUG_LOADER
	fprintf (stderr, "wav_close ()\n");
#endif
	if (priv->free)
		priv->free (sound_data);
	si_free (priv);
}

static Uint32
wav_decode_normal (SiSoundData * data)
{
	SiLoaderPrivate *priv = data->internal->loader_private;
	if (data->flags & SI_SOUND_DATA_STREAM)
	{
		Uint32 rest = priv->bytes_left;
		Uint32 read = data->internal->buffer_size;
		Uint32 num_read;

		if (read > rest)
			read = rest;
		
		num_read = si_read (data, data->internal->buffer, 1, read);

		rest -= read;

		BAIL_IO (num_read == read, data, num_read);
		
		if (rest == 0)
		{
			data->flags |= SI_SOUND_DATA_EOF;
		}
		
		priv->bytes_left = rest;
			
		return num_read;
	}
	else
	{
		int data_size;
		int read;
	
		data_size = priv->bytes_left;

		data->internal->buffer_size = data_size;
		data->internal->buffer = si_malloc (data_size);
		
		BAIL_MEM (data->internal->buffer != NULL, data, 0);

		read = si_read (data, data->internal->buffer, 1, data_size);

		BAIL_IO (read == data_size, data, 0);

		return read;
	}
}

/* 
 * MS ADPCM decompression 
 */

#define MS_ADPCM_SMALLEST_DELTA  16

static inline int 
ms_adpcm_read_block_headers (SiSoundData * data)
{
	SiLoaderPrivate *priv = data->internal->loader_private;
	MsAdpcmBlockHeader *headers = priv->ms_adpcm.block_headers;
	int i;
	int max = priv->format.channels;
	
	if (priv->bytes_left < priv->format.block_align)
	{
		data->flags |= SI_SOUND_DATA_EOF;
		return 0;
	}
	
	priv->bytes_left -= priv->format.block_align;
	
	for (i = 0; i < max; ++i)
		BAIL_IO (si_read_byte_ok (data, &headers [i].predictor), data, 0);
	for (i = 0; i < max; ++i)
		BAIL_IO (si_read_le16_ok (data, &headers [i].delta), data, 0);
	for (i = 0; i < max; ++i)
		BAIL_IO (si_read_le16_ok (data, &headers [i].samp1), data, 0);
	for (i = 0; i < max; ++i)
		BAIL_IO (si_read_le16_ok (data, &headers [i].samp2), data, 0);	

	priv->ms_adpcm.samples_left_in_block = priv->ms_adpcm.samples_per_block;
	priv->ms_adpcm.nibble_state = 0;
	
	return 1;
}


static inline void
ms_adpcm_do_nibble (Uint8 nib, MsAdpcmBlockHeader * header, Sint32 pred_samp)
{
	static const Sint32 max_audioval = ((1<<(16-1))-1);
	static const Sint32 min_audioval = -(1<<(16-1));
	static const Sint32 adaptation_table [] =
	{
		230, 230, 230, 230, 307, 409, 512, 614,
		768, 614, 512, 409, 307, 230, 230, 230
	};

	Sint32 new_samp;
	Sint32 delta;
	
	if (nib & 0x08)
		new_samp = pred_samp + (header->delta * (nib - 0x10));
	else
		new_samp = pred_samp + (header->delta * nib);
	
        /* clamp value... */
	if (new_samp < min_audioval)
		new_samp = min_audioval;
	else if (new_samp > max_audioval)
		new_samp = max_audioval;

	delta = ((Sint32) header->delta * adaptation_table [nib]) >> 8;

	if (delta < MS_ADPCM_SMALLEST_DELTA)
		delta = MS_ADPCM_SMALLEST_DELTA;
	
	header->delta = delta;
	header->samp2 = header->samp1;
	header->samp1 = new_samp;
}

static inline int 
ms_adpcm_decode_sample_frame (SiLoaderPrivate *priv)
{
	MsAdpcmBlockHeader *headers = priv->ms_adpcm.block_headers;
	Uint8 nib = priv->ms_adpcm.nibble;
	Sint16 coef1;
	Sint16 coef2;
	Sint32 pred_samp;

	int i;
	int max = priv->format.channels;

	for (i = 0; i < max; ++i)
	{
	
		coef1 = priv->ms_adpcm.coef [headers [i].predictor].coef1;
		coef2 = priv->ms_adpcm.coef [headers [i].predictor].coef2;
		pred_samp = ((headers [i].samp1 * coef1) +
		            (headers [i].samp2 * coef2)) >> 8;

		if (priv->ms_adpcm.nibble_state == 0)
		{
			nib = priv->ms_adpcm.block [priv->ms_adpcm.block_pos++];
			priv->ms_adpcm.nibble_state = 1;
			ms_adpcm_do_nibble (((Uint8) (nib >> 4)), (&headers [i]), (pred_samp));
		}
		else
		{
			priv->ms_adpcm.nibble_state = 0;
			ms_adpcm_do_nibble ((Uint8) (nib & 0x0F), (&headers [i]), (pred_samp));
		}
	} 

	priv->ms_adpcm.nibble = nib;
	return 1;	
}

static inline void 
ms_adpcm_put_frame1 (Uint16 *buf, SiLoaderPrivate * priv)
{
	static int i;
	for (i = 0; i < priv->format.channels; ++i)
   		*(buf)++ = priv->ms_adpcm.block_headers [i].samp1;
} 

static inline void 
ms_adpcm_put_frame2 (Uint16 *buf, SiLoaderPrivate * priv)
{
	static int i;
	for (i = 0; i < priv->format.channels; ++i)
   		*(buf)++ = priv->ms_adpcm.block_headers [i].samp2;
} 

static Uint32
wav_decode_ms_adpcm (SiSoundData * data)
{
	SiLoaderPrivate *priv = data->internal->loader_private;
	Uint32 bw = 0;
	Uint8 *buffer = data->internal->buffer;

	if ((data->flags & SI_SOUND_DATA_STREAM) == 0)
	{
		Uint32 buffer_size = priv->sample_frame_size * 
		                     priv->decompressed_size;
		buffer = si_malloc (buffer_size);	
		BAIL_MEM (buffer != NULL, data, 0);
		data->internal->buffer = buffer;
		data->internal->buffer_size = buffer_size;
	}

	while (bw < data->internal->buffer_size)
	{
#if 1	
		if (priv->ms_adpcm.samples_left_in_block > 1)
		{
			/* output latest sample frame and read a new one... */
			ms_adpcm_put_frame1 ((Uint16*) (buffer + bw), priv);
			-- priv->ms_adpcm.samples_left_in_block;
			bw += priv->sample_frame_size;

			if (!ms_adpcm_decode_sample_frame (priv))
			{
				data->flags |= SI_SOUND_DATA_ERROR;
				return bw;
			}
		}
		else if (priv->ms_adpcm.samples_left_in_block == 1)
		{
			/* output last sample frame of block... */
			ms_adpcm_put_frame1 ((Uint16*) (buffer + bw), priv);
			-- priv->ms_adpcm.samples_left_in_block;
			bw += priv->sample_frame_size;
		}
		else
		{
			/* need to read a new block... */

			if (!ms_adpcm_read_block_headers (data))
			{
				if (!(data->flags & SI_SOUND_DATA_EOF))
					data->flags |= SI_SOUND_DATA_ERROR;
				return 0;
			}
			
			/* only write first sample frame for now. */
			ms_adpcm_put_frame2 ((Uint16*) (buffer + bw), priv);
			-- priv->ms_adpcm.samples_left_in_block;
			bw += priv->sample_frame_size;
			priv->ms_adpcm.block_pos = 0;

			if (!si_read (data, priv->ms_adpcm.block,
			              priv->ms_adpcm.samples_left_in_block-1, 1))
			{
				si_set_error (SI_ERROR_IO);
				data->flags |= SI_SOUND_DATA_ERROR;
				return 0;
			}

		}

#else
		/* write ongoing sample frame before reading more data... */
		switch (priv->ms_adpcm.samples_left_in_block)
		{
		case 0:
			/* need to read a new block... */

			if (!ms_adpcm_read_block_headers (data))
			{
				if (!(data->flags & SI_SOUND_DATA_EOF))
				    data->flags |= SI_SOUND_DATA_ERROR;
				return 0;
			}
			
			/* only write first sample frame for now. */
			ms_adpcm_put_frame2 ((Uint16*) (buffer + bw), priv);
			-- priv->ms_adpcm.samples_left_in_block;
			bw += priv->sample_frame_size;
			priv->ms_adpcm.block_pos = 0;

			if (!si_read (data, priv->ms_adpcm.block,
			              priv->ms_adpcm.samples_left_in_block-1, 1))
			{
				si_set_error (SI_ERROR_IO);
				data->flags |= SI_SOUND_DATA_ERROR;
				return 0;
			}

			break;
		case 1:
			/* output last sample frame of block... */
			ms_adpcm_put_frame1 ((Uint16*) (buffer + bw), priv);
			-- priv->ms_adpcm.samples_left_in_block;
			bw += priv->sample_frame_size;
			break;
		default:
			/* output latest sample frame and read a new one... */
			ms_adpcm_put_frame1 ((Uint16*) (buffer + bw), priv);
			-- priv->ms_adpcm.samples_left_in_block;
			bw += priv->sample_frame_size;

			if (!ms_adpcm_decode_sample_frame (priv))
			{
				data->flags |= SI_SOUND_DATA_ERROR;
				return bw;
			}
			
		}
#endif
	}
	
	return bw;	
}

/*
 * IMA ADPCM decompression 
 */
 
static inline int
ima_adpcm_read_block_headers (SiSoundData * data)
{
	SiLoaderPrivate *priv = data->internal->loader_private;
	ImaAdpcmBlockHeader *headers = priv->ima_adpcm.block_headers;
	int i;
	int max = priv->format.channels;
	
	if (priv->bytes_left < priv->format.block_align)
	{
		data->flags |= SI_SOUND_DATA_EOF;
		return 0;
	}
	
	priv->bytes_left -= priv->format.block_align;
	
	for (i = 0; i < max; ++i)
	{
		Sint16 sample;
		BAIL_IO (si_read_le16_ok (data, &sample), data, 0);
		headers [i].sample = sample;
		BAIL_IO (si_read_byte_ok (data, &headers [i].index), data, 0);
		BAIL_IO (si_read_byte_ok (data, &headers [i].reserved), data, 0);
	}

	priv->ima_adpcm.samples_left_in_block = priv->ima_adpcm.samples_per_block;
	
	return 1;
}


static inline Sint32 
ima_adpcm_nibble(ImaAdpcmBlockHeader * header, Uint8 nibble)
{
	static const Sint32 max_audioval = ((1<<(16-1))-1);
	static const Sint32 min_audioval = -(1<<(16-1));
	static const int index_table [16] = 
	{
		-1, -1, -1, -1,
		 2,  4,  6,  8,
		-1, -1, -1, -1,
		 2,  4,  6,  8
	};
	static const Sint32 step_table [89] = 
	{
		7, 8, 9, 10, 11, 12, 13, 14, 16, 17, 19, 21, 23, 25, 28, 
		31, 34, 37, 41, 45, 50, 55, 60, 66, 73, 80, 88, 97, 107, 118,
		130, 143, 157, 173, 190, 209, 230, 253, 279, 307, 337, 371,
		408, 449, 494, 544, 598, 658, 724, 796, 876, 963, 1060, 1166,
		1282, 1411, 1552, 1707, 1878, 2066, 2272, 2499, 2749, 3024, 
		3327, 3660, 4026, 4428, 4871, 5358, 5894, 6484, 7132, 7845,
		8630, 9493, 10442, 11487, 12635, 13899, 15289, 16818, 18500, 
		20350, 22385, 24623, 27086, 29794, 32767
	};
	Sint32 delta, step;

	/* Compute difference and new sample value */
	step = step_table [header->index];
	delta = step >> 3;
	if (nibble & 0x04) delta += step;
	if (nibble & 0x02) delta += (step >> 1);
	if (nibble & 0x01) delta += (step >> 2);
	if (nibble & 0x08) delta = -delta;
	header->sample += delta;

	/* Update index value */
	header->index += index_table [nibble];
	if (header->index > 88) 
		header->index = 88;
	else if (header->index < 0)
		header->index = 0;

	/* Clamp output sample */
	if (header->sample > max_audioval)
		header->sample = max_audioval;
	else if (header->sample < min_audioval)
		header->sample = min_audioval;
	return header->sample;
}

/* Fill the decode buffer with a channel block of data (8 samples) */
static inline void 
ima_adpcm_fill_block (Uint8 * decoded, Uint8 * encoded, int channel, 
                      int num_channels, ImaAdpcmBlockHeader * header)
{
	int i;
	Sint8 nibble;
	Sint32 new_sample;
	decoded += (channel * 2);
	for (i = 0; i < 4; ++i)
	{
		nibble = (* encoded) & 0x0F;
		new_sample = ima_adpcm_nibble (header, nibble);
		* ((Uint16 *) decoded) = new_sample;
		decoded += 2 * num_channels;

		nibble = (* encoded) >> 4;
		new_sample = ima_adpcm_nibble (header, nibble);
		* ((Uint16 *) decoded) = new_sample;
		decoded += 2 * num_channels;
		++ encoded;
	}
}

static Uint32
wav_decode_ima_adpcm (SiSoundData * data)
{
	SiLoaderPrivate *priv = data->internal->loader_private;
	const int channels = priv->format.channels;
	Uint8 *buffer = data->internal->buffer;
	Uint8 *decoded = priv->ima_adpcm.decoded;
	ImaAdpcmBlockHeader *headers = priv->ima_adpcm.block_headers;
	Uint32 bw = 0;
	int i;

	if ((data->flags & SI_SOUND_DATA_STREAM) == 0)
	{
		Uint32 buffer_size = priv->sample_frame_size * 
		                     priv->decompressed_size;
		buffer = si_malloc (buffer_size);
		BAIL_MEM (buffer != NULL, data, 0);
		data->internal->buffer = buffer;
		data->internal->buffer_size = buffer_size;
	}

	while (bw < data->internal->buffer_size)
	{
		if (priv->ima_adpcm.samples_left_in_block == 0)
		{
			/* need to read a new block... */

			if (!ima_adpcm_read_block_headers (data))
			{
				if (!(data->flags & SI_SOUND_DATA_EOF))
					data->flags |= SI_SOUND_DATA_ERROR;
				return 0;
			}

			for (i = 0; i < channels; ++i)
			{
				*(Uint16 *) (buffer + bw + (2 * i)) 
				    = headers [i].sample;
			}
			bw += priv->sample_frame_size;
			-- priv->ima_adpcm.samples_left_in_block;
			continue;
		}
		else
		{
			if (priv->ima_adpcm.last_copied == 0)
			{
				Uint32 num = priv->ima_adpcm.decoded_size;
				
				for (i = 0; i < channels; ++i)
				{
					Uint8 encoded [4];
					BAIL_IO (si_read (data, encoded, 4, 1), 
					         data, 0);
					ima_adpcm_fill_block (decoded, encoded,
					                      i, channels, 
					                      &headers [i]);
				}
				if (num + bw >=
				    data->internal->buffer_size)
				{
					num = (data->internal->buffer_size - bw);
				}
				memcpy (buffer + bw, decoded, num);
				
				if (num != priv->ima_adpcm.decoded_size)
					priv->ima_adpcm.last_copied = num;
					
				bw += num;
				priv->ima_adpcm.samples_left_in_block -= 
					((num >> 1) / (channels));
			}
			else
			{
				int num_copy = priv->ima_adpcm.decoded_size
				               - priv->ima_adpcm.last_copied;
				memcpy (buffer + bw,
				        decoded + priv->ima_adpcm.last_copied,
				        num_copy);
				priv->ima_adpcm.last_copied = 0;
				bw += num_copy;
				priv->ima_adpcm.samples_left_in_block 
					-= ((num_copy >> 1) / channels);
				continue;
			}
		}
	};

	return bw;
}

static Uint32
wav_decode (SiSoundData * data)
{
	SiLoaderPrivate *priv = data->internal->loader_private;
#ifdef DEBUG_LOADER
	/* fprintf (stderr, "wav_decode ()\n"); */
#endif	
	if (priv->decode)
		return priv->decode (data);
	return 0;
}

static int
wav_rewind_normal (SiSoundData * sound_data)
{
	SiLoaderPrivate *priv = sound_data->internal->loader_private;

	BAIL_IO (wav_find_chunk (sound_data, "data", 1), sound_data, 0)
	BAIL_IO (si_read_le32_ok (sound_data, &priv->bytes_left), 
	         sound_data, 0);

	return 1;
}

static int
wav_rewind_ms_adpcm (SiSoundData * sound_data)
{
	SiLoaderPrivate *priv = sound_data->internal->loader_private;
	
	BAIL_IO (wav_find_chunk (sound_data, "data", 1), sound_data, 0)
	BAIL_IO (si_read_le32_ok (sound_data, &priv->bytes_left),
	         sound_data, 0);
	priv->ms_adpcm.samples_left_in_block = 0;
	priv->ms_adpcm.nibble_state = 0;
	priv->ms_adpcm.nibble = 0;
	return 1;
}

static int
wav_rewind_ima_adpcm (SiSoundData * sound_data)
{
	SiLoaderPrivate *priv = sound_data->internal->loader_private;

	BAIL_IO (wav_find_chunk (sound_data, "data", 1), sound_data, 0)
	BAIL_IO (si_read_le32_ok (sound_data, &priv->bytes_left),
	         sound_data, 0);
	priv->ima_adpcm.samples_left_in_block = 0;
	priv->ima_adpcm.last_copied = 0;
	return 1;
}

static int
wav_rewind (SiSoundData * data)
{
	SiLoaderPrivate *priv = data->internal->loader_private;
#ifdef DEBUG_LOADER
	fprintf (stderr, "wav_rewind ()\n"); 
#endif	
	if (priv->rewind)
		return priv->rewind (data);
	return 0;
}

SiLoaderInternal _si_wav_internal =
{
	wav_init,
	wav_quit,
	wav_open,
	wav_close,
	wav_decode,
	wav_rewind
};

