/* 
 * SI - Sound Interface
 * Copyright (c) 2002 Matej Knopp
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <SDL/SDL.h>

#define SI_LITTLE_ENDIAN 1234
#define SI_BIG_ENDIAN    4321

/* nothing since we are not under win32 */
#define SI_DECLSPEC DECLSPEC

#ifndef __cplusplus
#define inline __inline
#endif

/*
 * Basic types and endian type
 */

#ifndef __si_config_h__
#define __si_config_h__

#ifndef _SDL_types_h

#include <stdint.h> /* ISO C99 */

typedef uint8_t  Uint8;
typedef uint16_t Uint16;
typedef uint32_t Uint32;

typedef int8_t  Sint8;
typedef int16_t Sint16;
typedef int32_t Sint32;

#include <endian.h>

#if __BYTE_ORDER == __LITTLE_ENDIAN
#  define SI_BYTE_ORDER SI_LITTLE_ENDIAN
#else
#  define SI_BYTE_ORDER SI_BIG_ENDIAN
#endif

#else /* _SDL_types_h */

#define SI_BYTE_ORDER    SDL_BYTEORDER

#endif /* _SDL_types_h */

/*
 * Memory
 */

/* 
 * Very simple memory usage tracing.
 * Counts the number of allocated buffers.
 * Not thread-safe.
 */

#define SI_MEM_DBG_NONE    0
#define SI_MEM_DBG_NORMAL  1
#define SI_MEM_DBG_VERBOSE 2

#define SI_MEM_DBG SI_MEM_DBG_NORMAL

#include "malloc.h"

#if SI_MEM_DBG == SI_MEM_DBG_NONE
#  define si_malloc(arg)        malloc ((arg))
#  define si_realloc(arg,size)  realloc ((arg), (size))
#  define si_free(arg)          free ((arg))
#  define si_malloc0(arg)       calloc (1, (arg))

#else /* SI_MEM_DBG == SI_MEM_DBG_NONE */

#ifdef _MSC_VER
#  define __mark __FILE__
#else
#  define __mark __FILE__ " ["  __func__ "]"
#endif
#  define si_malloc(arg)        si_malloc_dbg ((arg), __mark, __LINE__)
#  define si_realloc(arg,size)  si_realloc_dbg ((arg), (size), #arg, __mark, __LINE__)
#  define si_free(arg)          si_free_dbg ((arg), #arg, __mark, __LINE__)
#  define si_malloc0(arg)       si_malloc0_dbg ((arg), __mark, __LINE__)

#include <stdio.h>

/* !FIXME! - This is NOT thread safe */

extern int si_num_allocated;

static inline void *
si_malloc_dbg (size_t size, const char * where, int line)
{
#if SI_MEM_DBG == SI_MEM_DBG_VERBOSE
	fprintf (stderr, "+ malloc  () %-8i %s on line %i\n", size, where, line);
#endif /* SI_MEM_DBG == SI_MEM_DBG_VERBOSE */
	si_num_allocated ++;
	return malloc (size);
}

static inline void *
si_malloc0_dbg (size_t size, const char * where, int line)
{
#if SI_MEM_DBG == SI_MEM_DBG_VERBOSE
	fprintf (stderr, "+ malloc0 () %-8i %s on line %i\n", size, where, line);
#endif /* SI_MEM_DBG == SI_MEM_DBG_VERBOSE */
	si_num_allocated ++;
	return calloc (1, size);
}

static inline void *
si_realloc_dbg (void * arg, size_t size, const char * what, 
                const char * where, int line)
{
#if SI_MEM_DBG == SI_MEM_DBG_VERBOSE
	fprintf (stderr, "+ realloc () %-8i %s on line %i  (%s)\n", size, where, line, what);
#endif /* SI_MEM_DBG == SI_MEM_DBG_VERBOSE */
	return realloc (arg, size);
}


static inline void
si_free_dbg (void * ptr, const char * what, const char * where, int line)
{
#if SI_MEM_DBG == SI_MEM_DBG_VERBOSE
	fprintf (stderr, "- free    ()          %s on line %i  (%s)\n", where, line, what);
#endif /* SI_MEM_DBG == SI_MEM_DBG_VERBOSE */
	si_num_allocated --;
	free (ptr);
}

#endif /* SI_MEM_DBG != SI_MEM_DBG_NONE */

#define si_new(struct_type, n_structs)		\
    ((struct_type *) si_malloc (((size_t) sizeof (struct_type)) * ((size_t) (n_structs))))

#define si_renew(struct_type, mem, n_structs)	\
    ((struct_type *) si_realloc ((mem), ((size_t) sizeof (struct_type)) * ((size_t) (n_structs))))

#define si_new0(struct_type, n_structs)	\
    ((struct_type *) si_malloc0 (((size_t) sizeof (struct_type)) * ((size_t) (n_structs))))


/*
 * Debug
 */

#ifdef SI_DEBUG

#include <stdio.h>

#define si_return_val_if_fail(arg, val) \
	{ if (!(arg)) \
	    { \
	       fprintf (stderr, "WARNING: %s:%i - in function `%s'\n         - assertion `%s' failed\n", \
	    	        __FILE__, __LINE__, __PRETTY_FUNCTION__, #arg); \
	       return (val); \
	    } \
	}

#define si_return_if_fail(arg) \
	{ if (!(arg)) \
	    { \
	       fprintf (stderr, "WARNING: %s:%i - in function `%s'\n         - assertion `%s' failed\n", \
	    	        __FILE__, __LINE__, __PRETTY_FUNCTION__, #arg); \
	       return; \
	    } \
	}

#else /* SI_DEBUG */

#define si_return_val_if_fail(arg, val) if (!(arg)) return (val);
#define si_return_if_fail(arg) if (!(arg)) return;

#endif /* SI_DEBUG */

/*
 * Endiannes conversion (from SDL)
 */
 
#ifdef _SDL_endian_h

#define si_swap16 SDL_Swap16
#define si_swap32 SDL_Swap32

#define si_swap_le16 SDL_SwapLE16
#define si_swap_le32 SDL_SwapLE32
#define si_swap_be16 SDL_SwapBE16
#define si_swap_be32 SDL_SwapBE32

#else
 
/* The macros used to swap values */
/* Try to use superfast macros on systems that support them */
#ifdef linux
#include <asm/byteorder.h>
#ifdef __arch__swab16
#define si_swap16  __arch__swab16
#endif
#ifdef __arch__swab32
#define si_swap32  __arch__swab32
#endif
#endif /* linux */
/* Use inline functions for compilers that support them, and static
   functions for those that do not.  Because these functions become
   static for compilers that do not support inline functions, this
   header should only be included in files that actually use them.
*/
#ifndef si_swap16
static __inline__ Uint16 si_swap16(Uint16 D) {
	return((D<<8)|(D>>8));
}
#endif
#ifndef si_swap32
static __inline__ Uint32 si_swap32(Uint32 D) {
	return((D<<24)|((D<<8)&0x00FF0000)|((D>>8)&0x0000FF00)|(D>>24));
}
#endif

/* Byteswap item from the specified endianness to the native endianness */
#if SI_BYTE_ORDER == SI_LITTLE_ENDIAN
#define si_swap_le16(X) (X)
#define si_swap_le32(X) (X)
#define si_swap_be16(X) si_swap16(X)
#define si_swap_be32(X) si_swap32(X)
#else
#define si_swap_le16(X) si_swap16(X)
#define si_swap_le32(X) si_swap32(X)
#define si_swap_be16(X) (X)
#define si_swap_be32(X) (X)
#endif

#endif /* _SDL_endian_h */

#endif /* __si_config_h__ */

