/* 
 * SI - Sound Interface
 * Copyright (c) 2002 Matej Knopp
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifndef __si_internal_h__
#define __si_internal_h__

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

/*
 * Audio conversion
 */

typedef struct _SiAudioConvertor SiAudioConvertor;

/* A structure to hold a set of audio conversion filters and buffers */
struct _SiAudioConvertor  {
	int      needed;        /* Set to 1 if conversion possible */
	Uint16   src_format;    /* Source audio format */
	Uint16   dst_format;    /* Target audio format */
	double   rate_incr;     /* Rate conversion increment */
	Uint8  * buf;           /* Buffer to hold entire audio data */
	int      len;           /* Length of original audio buffer */
	int      len_cvt;       /* Length of converted audio buffer */
	int      len_mult;      /* buffer must be len * len_mult big */
	double   len_ratio;     /* Given len, final size is 
	                         * len * len_ratio */
	void (* filters[10]) (SiAudioConvertor * cvt, Uint16 format);
	int filter_index;       /* Current audio conversion function */
};

/* 
 * Creates a set of audio filters to convert from one format to another. 
 * Returns 0 if the format conversion is not supported, or 1 if the
 * audio filter is set up.
 */

extern SI_DECLSPEC
int si_build_audio_convertor (SiAudioConvertor * cvt, Uint16 src_format,
                              Uint8 src_channels, int src_rate,
                              Uint16 dst_format, Uint8 dst_channels, 
                              int dst_rate);


/*
 * Returns 0, if any error has occured. Otherwise returns 1.
 */

extern SI_DECLSPEC
int si_convert_audio (SiAudioConvertor * cvt);

/*
 * Internal
 */

struct _SiLoaderInternal
{
	int    (* init_loader) (void);
	void   (* quit_loader) (void);
	int    (* open)        (SiSoundData * sound_data, 
	                        const char * extension,
	                        const SiLoaderArgument * arguments);
	void   (* close)       (SiSoundData * sound_data);
	Uint32 (* decode)      (SiSoundData * sound_data);
	int    (* rewind)      (SiSoundData * sound_data);

};

/* Returns newly allocated list of available loaders' ids. You should
 * free the result with si_free (but not the items themselves.) */
extern SI_DECLSPEC
const char ** si_get_loaders ();

extern SI_DECLSPEC
int si_register_user_loader (SiLoaderInfo * info,
                             SiLoaderInternal * internal);

typedef struct _SiLoaderPrivate SiLoaderPrivate;

struct _SiSoundDataInternal
{
	SiAudioInfo        actual;         /* the format decoder is using */
	Uint8            * buffer;         /* the internal buffer */
	Uint32             buffer_size;    /* the internal buffer size */
	SiAudioConvertor   convertor;
	SiSoundDataInput   input;
	void             * input_context;
	SiLoaderPrivate  * loader_private; /* allocated and freed by loader */
	Uint32             data1;          /* used by loader */
	Uint32             data2;          /* used by loader */
	void             * link;           /* private */
};

/*
 * Buffer policy
 * -------------
 * The Sound data contains two buffers. SiSoundData->buffer and
 * SiSoundData->internal->buffer. The first contains data in
 * the SiSoundData->requested format. The second contains data in
 * the internal format used by the decoder. If the formats are equal
 * the buffers are shared. (The internal->buffer is a pointer to 
 * SiSoundData->buffer. 
 *
 * Allocation of buffers
 * ---------------------
 * When and how are buffers allocated depends on whether the sound
 * data is streamable or not. 
 * a) If the data is streamable, both buffers are created within 
 *    the si_sound_data_new_stream () function. 
 *    The loader->internal->open () function must set the
 *    SiSoundData->internal->actual format and the buffer will
 *    be created according to it.
 * b) If the data is not streamable, the internal buffer should be 
 *    created by the loader->internal->decode () function and the
 *    "public" buffer will be created automatically after it.
 */

/*
 * Errors
 */

#define SI_ERROR_MEM     "Out of memory."
#define SI_ERROR_IO      "I/O error."

extern SI_DECLSPEC
void si_set_error (const char * error);

/*
 * Case insensitive string conversion function.
 */

extern SI_DECLSPEC
int si_strcasecmp (const char * str1, const char * str2);

/* Helper functions */

extern SI_DECLSPEC Uint8  si_read_byte (SiSoundData * data);
extern SI_DECLSPEC Uint16 si_read_le16 (SiSoundData * data);
extern SI_DECLSPEC Uint32 si_read_le32 (SiSoundData * data);
extern SI_DECLSPEC Uint16 si_read_be16 (SiSoundData * data);
extern SI_DECLSPEC Uint32 si_read_be32 (SiSoundData * data);

extern SI_DECLSPEC int si_read_byte_ok (SiSoundData * data, Uint8 * dst);
extern SI_DECLSPEC int si_read_le16_ok (SiSoundData * data, Uint16 * dst);
extern SI_DECLSPEC int si_read_le32_ok (SiSoundData * data, Uint32 * dst);
extern SI_DECLSPEC int si_read_be16_ok (SiSoundData * data, Uint16 * dst);
extern SI_DECLSPEC int si_read_be32_ok (SiSoundData * data, Uint32 * dst);

#define si_seek(data, offset, whence)    data->internal->input.seek (data->internal->input_context, offset, whence)
#define si_read(data, ptr, size, maxnum) data->internal->input.read (data->internal->input_context, ptr, size, maxnum)
#define si_close_input(data)             do { data->internal->input.close (data->internal->input_context); data->internal->input_context = NULL; } while (0)

#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* __si_internal_h__ */


