/* 
 * SI - Sound Interface
 * Copyright (c) 2002 Matej Knopp
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */
 
#include "si.h"
#include "si-internal.h"
#include "gem-list.h"

#include <string.h>
#include <stdarg.h>
#include <stdio.h>
#include <ctype.h>

#define ERR_MAX_LEN 256 
static char _error [ERR_MAX_LEN + 1] = { 0 };

#if SI_MEM_DBG != SI_MEM_DBG_NONE
int si_num_allocated = 0;
#endif

void 
si_set_error (const char * error)
{
	strncpy (_error, error, ERR_MAX_LEN);
}

const char *
si_get_error (void)
{
	return _error;
}

static GemList *loaders = NULL;

static GemList *sound_datas = NULL;

#define SI_LOADER(arg)  ((SiLoader *)arg)

SiLoader *
si_get_loader (const char * id)
{
	GemList *list = loaders;

	si_return_val_if_fail (id != NULL, NULL);
	
	while (list)
	{
		if (list->data && 
		    strcmp (SI_LOADER (list->data)->info->id, id) == 0)
		{
			return SI_LOADER (list->data);
		}
		list = gem_list_next (list);
	}
	
	return NULL;
}

SiLoader *
si_find_loader (const char * ext)
{
	GemList *list = loaders;

	si_return_val_if_fail (ext != NULL, NULL);

	while (list)
	{
		if (list->data && SI_LOADER (list->data)->info->extensions)
		{
			SiLoader *tmp;
			const char *e;
			int i = 0;
			
			tmp = SI_LOADER (list->data);
			e = tmp->info->extensions [0];
			while (e)
			{
				if (si_strcasecmp (e, ext) == 0)
					return tmp;

				e = tmp->info->extensions [++i];
			}
		}
		list = gem_list_next (list);
	}
	
	return NULL;
}

SiLoader *
si_prefer_loader (const char * id)
{
	GemList *list = loaders;

	si_return_val_if_fail (id != NULL, NULL);

	while (list)
	{
		if (list->data && 
		    strcmp (SI_LOADER (list->data)->info->id, id) == 0)
		{
			if (list != loaders) 
			{
				void *tmp = list->data;
				list->data = loaders->data;
				loaders->data = tmp;
			}
			return SI_LOADER (loaders->data);
		}
		list = gem_list_next (list);
	}
	
	return NULL;
};

static int
si_register_loader (SiLoaderInfo * info, SiLoaderInternal * internal,
                    SiLoaderType type)
{
	SiLoader *loader;

	si_return_val_if_fail (info != NULL, 0);
	si_return_val_if_fail (internal != NULL, 0);

	if (si_find_loader (info->id) != NULL)
		return 0;
		
	loader = si_new0 (SiLoader, 1);
	if (!loader)
	{
		si_set_error (SI_ERROR_MEM);
		return 0;
	}
		
	loader->info = info;
	loader->internal = internal;
	loader->type = type;
		
	loaders = gem_list_prepend (loaders, loader);
	
	return 1;
}

const char **
si_get_loaders ()
{
	const char ** res;
	int i = 0;
	Uint32 num_loaders = gem_list_length (loaders);
	GemList *list;
	
	if (num_loaders == 0)
		return NULL;

	res = si_malloc (sizeof (const char *) * num_loaders);
	list = loaders;
	while (list)
	{
		res [i] =  SI_LOADER (list->data)->info->id;
		list = gem_list_next (list);
		++ i;
	}
	return res;
}

int 
si_register_user_loader (SiLoaderInfo * info, 
                         SiLoaderInternal * internal)
{
	si_return_val_if_fail (info != NULL, 0);
	si_return_val_if_fail (internal != NULL, 0);

	if (!internal->init_loader ())
		return 0;

	if (!si_register_loader (info, internal, SI_LOADER_TYPE_USER))
		return 0;

	return 1;
}

static int 
si_register_builtin_loader (SiLoaderInfo * info, 
                            SiLoaderInternal * internal)
{
	si_return_val_if_fail (info != NULL, 0);
	si_return_val_if_fail (internal != NULL, 0);

	if (!internal->init_loader ())
		return 0;

	if (!si_register_loader (info, internal, SI_LOADER_TYPE_BUILTIN))
		return 0;

	return 1;
}


static int
si_audio_info_equals (SiAudioInfo *i1, SiAudioInfo *i2)
{
	if (!i1 || !i2)
		return 0;
	return (i1->rate == i2->rate && i1->format == i2->format &&
		i1->channels == i2->channels);
}

static SiLoaderArgument *
get_arguments (const char * arg1, va_list _args)
{
	const char *s = arg1;
	int count = 0;
	int i;
	SiLoaderArgument *res;
	va_list args;

	s = arg1;

	args = _args;
	while (s)
	{
		s = va_arg (args, const char *);
		++count;
	}

	if (count % 2) /* we need pairs ! */
		count--;

	count /= 2;

	res = si_new (SiLoaderArgument, count + 1);

	s = arg1;
	args = _args;
	for (i = 0; i < count; i++)
	{
		res [i].key = s;
		s = va_arg (args, const char *);
		res [i].value = s;
		s = va_arg (args, const char *);
	}
	
	res [count].key = res [count].value = NULL;
	
	return res;
};

static SiLoader *
find_loader (SiSoundData * data, const char * extension,
             const SiLoaderArgument * arguments)
{
	SiLoader *ext = NULL;
	GemList  *list;
	if (extension)
	{
		ext = si_find_loader (extension);
		if (ext && ext->internal->open (data, extension, arguments))
			return ext;
	}
	list = loaders;
	while (list)
	{
		SiLoader *l = SI_LOADER (list->data);
		si_seek (data, 0, SEEK_SET);
		if (l && l != ext && 
		    l->internal->open (data, extension, arguments))
			return l;
		
		list = gem_list_next (list);
	}
	return NULL;
};

SiSoundData * 
si_sound_data_new (const SiSoundDataInput * input, void * input_context,
                   const SiAudioInfo * requested, const char * extension, 
                   const char * arg1, ...)
{
	SiSoundData      *res;
	SiLoaderArgument *arguments;
	va_list           args;

	va_start (args, arg1);
	arguments = get_arguments (arg1, args);
	va_end (args);

	res = si_sound_data_new_arg (input, input_context, requested,
	                             extension, arguments);

	si_free (arguments);
	
	return res;
}		   

SiSoundData * 
si_sound_data_new_arg (const SiSoundDataInput * input, void * input_context,
                       const SiAudioInfo * requested, const char * extension, 
                       const SiLoaderArgument * args)
{
	SiSoundData         *sound_data;
	SiSoundDataInternal *sound_data_internal;
	SiLoader            *loader;
	
	si_return_val_if_fail (input != NULL, NULL);
	si_return_val_if_fail (requested != NULL, NULL);

	sound_data = si_new0 (SiSoundData, 1);
	sound_data_internal = si_new0 (SiSoundDataInternal, 1);
	if (!sound_data || !sound_data_internal)
	{
		if (sound_data)
			si_free (sound_data);
		input->close (input_context);
		si_set_error (SI_ERROR_MEM);
		return NULL;
	}	
	
	sound_data->requested = *requested;
	sound_data->internal = sound_data_internal;
	sound_data_internal->input = *input;
	sound_data_internal->input_context = input_context;

	loader = find_loader (sound_data, extension, args);
	
	if (!loader)
	{
		si_free (sound_data->internal);
		si_free (sound_data);
		input->close (input_context);
		if (!si_get_error () [0])
			si_set_error ("Couldn't find appropriate loader.");
		return NULL;
	}
	
	sound_data->loader = loader;

	sound_datas = gem_list_prepend (sound_datas, sound_data);
	sound_data->internal->link = sound_datas;
	
	return sound_data;
}


static void
sound_data_close_input (SiSoundData * sound_data)
{
	if (sound_data->internal->input_context)
	{
		sound_data->internal->input.close 
			(sound_data->internal->input_context);
	}

}

SiSoundData * 
si_sound_data_new_stream (const SiSoundDataInput * input, void * input_context,
                          const SiAudioInfo * requested, Uint32 buffer_size, 
                          const char * extension, const char * arg1, ...)
{
	SiLoaderArgument *arguments;
	va_list           args;
	SiSoundData      *res;

	va_start (args, arg1);
	arguments = get_arguments (arg1, args);
	va_end (args);

	res = si_sound_data_new_stream_arg (input, input_context, requested,
	                                    buffer_size, extension, arguments);

	si_free (arguments);
	
	return res;
}

SiSoundData * 
si_sound_data_new_stream_arg (const SiSoundDataInput * input, 
                              void * input_context, 
                              const SiAudioInfo * requested, 
                              Uint32 buffer_size, 
                              const char * extension, 
                              const SiLoaderArgument * args)
{
	SiSoundData         *sound_data;
	SiSoundDataInternal *sound_data_internal;
	SiLoader            *loader;
	
	si_return_val_if_fail (input != NULL, NULL);
	si_return_val_if_fail (requested != NULL, NULL);
	
	sound_data = si_new0 (SiSoundData, 1);
	sound_data_internal = si_new0 (SiSoundDataInternal, 1);
	if (!sound_data || !sound_data_internal)
	{
		if (sound_data)
			si_free (sound_data);
		input->close (input_context);
		si_set_error (SI_ERROR_MEM);
		return NULL;
	}
	
	sound_data->requested = *requested;
	sound_data->flags |= SI_SOUND_DATA_STREAM;
	sound_data->internal = sound_data_internal;
	sound_data_internal->input = *input;
	sound_data_internal->input_context = input_context;

	sound_data->buffer_size = buffer_size;
	sound_data->buffer = si_malloc0 (buffer_size);
	if (!sound_data->buffer)
	{	
		si_free (sound_data->internal);
		si_free (sound_data);
		input->close (input_context);
		si_set_error (SI_ERROR_MEM);
		return NULL;
	}

	loader = find_loader (sound_data, extension, args);
	
	if (!loader)
	{
		si_free (sound_data->buffer);
		si_free (sound_data->internal);
		si_free (sound_data);
		input->close (input_context);
		si_set_error ("Coudln't find appropriate loader.");
		return NULL;
	}
	
	/* allocate the buffer (if needed) */
	if (si_audio_info_equals (&sound_data->requested,
	                          &sound_data->internal->actual))
	{
		sound_data->internal->buffer = sound_data->buffer;
		sound_data->internal->buffer_size = sound_data->buffer_size;
	}
	else
	{
		si_build_audio_convertor 
			(&sound_data->internal->convertor,
			 sound_data->internal->actual.format,
			 sound_data->internal->actual.channels,
			 sound_data->internal->actual.rate,
			 sound_data->requested.format,
			 sound_data->requested.channels,
			 sound_data->requested.rate);

		/* compute new buffer size */
		sound_data->internal->buffer_size = (Uint32)
			(sound_data->buffer_size 
			  / sound_data->internal->convertor.len_ratio);

		sound_data->internal->buffer = 
			si_malloc0 (sound_data->internal->buffer_size);

		if (!sound_data->internal->buffer)
		{
			sound_data->loader->internal->close (sound_data);
			si_free (sound_data->buffer);
			si_free (sound_data->internal);
			si_free (sound_data);
			input->close (input_context);
			si_set_error (SI_ERROR_MEM);
			return NULL;
		}		
	}
	
	sound_data->loader = loader;

	sound_datas = gem_list_prepend (sound_datas, sound_data);
	sound_data->internal->link = sound_datas;
	
	return sound_data;
}

void 
si_sound_data_free (SiSoundData * sound_data)
{
	GemList *link;	

	si_return_if_fail (sound_data != NULL);
	
	sound_data->loader->internal->close (sound_data);	

	sound_data_close_input (sound_data);		

	if (sound_data->internal->buffer && 
	    sound_data->internal->buffer != sound_data->buffer)
	{
		si_free (sound_data->internal->buffer);
	}

	link = (GemList *) sound_data->internal->link;
	sound_datas = gem_list_delete_node (sound_datas, link);

	si_free (sound_data->internal);	

	if (sound_data->buffer)
		si_free (sound_data->buffer);

	si_free (sound_data);	
}

Uint32
si_sound_data_decode (SiSoundData * sound_data)
{
	Uint32 res;

	si_return_val_if_fail (sound_data != NULL, -1);
	
	if (sound_data->flags & SI_SOUND_DATA_EOF)
	{
		return 0;
	}
	
	res = sound_data->loader->internal->decode (sound_data);

	if (sound_data->flags & SI_SOUND_DATA_ERROR)
	{
		if (!si_get_error () [0])
			si_set_error ("Unknown sound decoding error.");
		return 0;
	}
	
	if (sound_data->flags & SI_SOUND_DATA_STREAM)
	{
		Uint32 res2;
		Uint8 *cvt_buf = sound_data->internal->buffer;

		if (!res)
			return 0;
			
		if (sound_data->buffer == sound_data->internal->buffer)
			return res;
			
		res2 = (Uint32) (res * sound_data->internal->convertor.len_ratio);
		
		if (res2 > sound_data->internal->buffer_size)
		{
			cvt_buf = sound_data->buffer;
			memcpy (sound_data->buffer,
				sound_data->internal->buffer, res);			
		}

		sound_data->internal->convertor.buf = cvt_buf;
		sound_data->internal->convertor.len = res;
		
		si_convert_audio (&sound_data->internal->convertor);

		if (cvt_buf == sound_data->internal->buffer)
		{
			memcpy (sound_data->buffer,
				sound_data->internal->buffer, res2);
		}

		return res2;
	}
	else
	{
		/* we no longer need the data */
		si_close_input (sound_data);
		
		if (si_audio_info_equals (&sound_data->requested,
					  &sound_data->internal->actual))
		{
			sound_data->buffer = sound_data->internal->buffer;
			sound_data->buffer_size = 
				sound_data->internal->buffer_size;
			return res;
		}
		else
		{
			Uint8 *cvt_buf = sound_data->internal->buffer;

			si_build_audio_convertor 
				(&sound_data->internal->convertor,
				 sound_data->internal->actual.format,
				 sound_data->internal->actual.channels,
				 sound_data->internal->actual.rate,
				 sound_data->requested.format,
				 sound_data->requested.channels,
				 sound_data->requested.rate);
			
			sound_data->buffer_size = (Uint32)
				(sound_data->internal->buffer_size 
				  * sound_data->internal->convertor.len_ratio);
			
			sound_data->buffer = 
				si_malloc (sound_data->buffer_size);

			if (!sound_data->buffer)
			{
				si_set_error (SI_ERROR_MEM);
				return -1;
			}
			
			/* convert the audio */
			
			if (sound_data->buffer_size > 
			    sound_data->internal->buffer_size)
			{
				memcpy (sound_data->buffer,
				        sound_data->internal->buffer,
				        sound_data->internal->buffer_size);
				cvt_buf = sound_data->buffer;
			}
			
			sound_data->internal->convertor.buf = cvt_buf;
			sound_data->internal->convertor.len = 
				sound_data->internal->buffer_size;
			
			si_convert_audio (&sound_data->internal->convertor);

			if (cvt_buf == sound_data->internal->buffer)
			{
				memcpy (sound_data->buffer,
				        sound_data->internal->buffer,
				        sound_data->buffer_size);
			}
			
			si_free (sound_data->internal->buffer);
			sound_data->internal->buffer = NULL;
			
			return sound_data->buffer_size;
		}
	};
}

int
si_sound_data_rewind (SiSoundData * sound_data)
{
	si_return_val_if_fail (sound_data != NULL, -1);
	si_return_val_if_fail (sound_data->flags & SI_SOUND_DATA_STREAM, -1);
	
	return sound_data->loader->internal->rewind (sound_data);
}

extern SiLoaderInfo     _si_raw_info;
extern SiLoaderInternal _si_raw_internal;
extern SiLoaderInfo     _si_wav_info;
extern SiLoaderInternal _si_wav_internal;

static void
si_register_loaders ()
{
	si_register_builtin_loader (&_si_raw_info, &_si_raw_internal);
	si_register_builtin_loader (&_si_wav_info, &_si_wav_internal);
}

static int num_init = 0;

int
si_init (void)
{
	if (num_init != 0)
	{
		++ num_init;
		return 1;
	}

	si_register_loaders ();

	if (!loaders) /* we couldn't register any loader! */
	{
		si_set_error ("Couldn't register any built-in loader!");
		return 0;
	}

	++ num_init;

	return 1;
}

void 
si_quit (void)
{
	if (num_init == 0)
		return;
	-- num_init;
	if (num_init > 0)
		return;

	/* delete existing datas */
	while (sound_datas)
	{
		if (sound_datas->data)
		{
			si_sound_data_free ((SiSoundData *)sound_datas->data);
		}
		sound_datas = gem_list_delete_node (sound_datas, sound_datas);
	}

	/* delete existing loaders */
	while (loaders)
	{
		if (loaders->data)
		{
			SI_LOADER (loaders->data)->internal->quit_loader ();
			si_free (loaders->data);			
		}
		loaders = gem_list_delete_node (loaders, loaders);
	}
}

int 
si_strcasecmp (const char * str1, const char * str2)
{
	int res;
	
	if (str1 == str2)
		return 0;
	if (!str1)
		return -1;
	if (!str2)
		return 1;
		
	while (*str1 || *str2)
	{
		res = toupper ((int) *str1) - toupper ((int) *str2);
		if (res != 0)
			return res;
		++str1;
		++str2;
	}
	
	return 0;
}

Uint8
si_read_byte (SiSoundData * data)
{
	Uint8 value;
	si_read (data, &value, sizeof (value), 1);
	return value;
}

Uint16 
si_read_le16 (SiSoundData * data)
{
	Uint16 value;
	si_read (data, &value, sizeof (value), 1);
	return si_swap_le16 (value);
}

Uint32 
si_read_le32 (SiSoundData * data)
{
	Uint32 value;
	si_read (data, &value, sizeof (value), 1);
	return si_swap_le32 (value);
}

Uint16 
si_read_be16 (SiSoundData * data)
{
	Uint16 value;
	si_read (data, &value, sizeof (value), 1);
	return si_swap_be16 (value);
}

Uint32 
si_read_be32 (SiSoundData * data)
{
	Uint32 value;
	si_read (data, &value, sizeof (value), 1);
	return si_swap_be32 (value);
}

int
si_read_byte_ok (SiSoundData * data, Uint8 * dst)
{
	return si_read (data, dst, 1, 1);
}

int 
si_read_le16_ok (SiSoundData * data, Uint16 * dst)
{
	int res;
	res = si_read (data, dst, 2, 1);
	*dst = si_swap_le16 (*dst);
	return res;
}

int 
si_read_le32_ok (SiSoundData * data, Uint32 * dst)
{
	int res;
	res = si_read (data, dst, 4, 1);
	*dst = si_swap_le32 (*dst);
	return res;
}

int 
si_read_be16_ok (SiSoundData * data, Uint16 * dst)
{
	int res;
	res = si_read (data, dst, 2, 1);
	*dst = si_swap_be16 (*dst);
	return res;
}

int 
si_read_be32_ok (SiSoundData * data, Uint32 * dst)
{
	int res;
	res = si_read (data, dst, 4, 1);
	*dst = si_swap_be32 (*dst);
	return res;
}


