/* 
 * SI - Sound Interface
 * Copyright (c) 2002 Matej Knopp
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */
 
#ifndef __si_h__
#define __si_h__

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

#include "si-config.h"

/*
 * Loader stuff.
 */

typedef struct _SiLoader            SiLoader;
typedef struct _SiLoaderInfo        SiLoaderInfo;
typedef struct _SiLoaderArgument    SiLoaderArgument;
typedef struct _SiAudioInfo         SiAudioInfo;
typedef struct _SiSoundDataInput    SiSoundDataInput;
typedef struct _SiSoundData         SiSoundData;

typedef struct _SiLoaderInternal    SiLoaderInternal;
typedef struct _SiSoundDataInternal SiSoundDataInternal;

enum _SiSoundDataFlags
{
	SI_SOUND_DATA_STREAM    = 1 << 0, /* Sound data is streamable. 
	                                     This flag is set by the
	                                     sound_data_new_stream func. */
	SI_SOUND_DATA_EOF       = 1 << 1, /* The input has reached it's end.
	                                     This flag is set by the loader. */
	SI_SOUND_DATA_ERROR     = 1 << 2  /* An error has occured... 
	                                     This flag is set by the loader*/
};

enum _SiLoaderType 
{
	SI_LOADER_TYPE_BUILTIN, /* builtin (registered within si_init ()) */
	SI_LOADER_TYPE_USER,    /* user registered (after si_init ()) */
	SI_LOADER_TYPE_PLUGIN   /* not yet supported */
};

typedef enum   _SiSoundDataFlags SiSoundDataFlags;
typedef enum   _SiLoaderType     SiLoaderType;

struct _SiLoaderInfo
{
	const char  * id;          /* no two loaders can have the same id */
	const char ** extensions;  /* array of supported extensions
	                              (last must be NULL) */
	const char  * description; /* short description of loader */
	const char  * author;      /* author (and his e-mail) of loader */
	const char  * url;         /* for more information about format, ... */
};

struct _SiLoaderArgument
{
	const char * key;
	const char * value;
};


struct _SiLoader
{
	SiLoaderInfo     * info;
	SiLoaderType       type;
	SiLoaderInternal * internal;
};

/* return loader with given id (or NULL, if no loader with 
 * given id is registered) */
extern SI_DECLSPEC
SiLoader * si_get_loader (const char * id);

/* return the first loader which supports files with given extension, or
 * if none */
extern SI_DECLSPEC
SiLoader * si_find_loader (const char * ext);

/* give the loader with given id higher priority */
extern SI_DECLSPEC
SiLoader * si_prefer_loader (const char * id);

/*
 * Sound data stuff.
 */

/*
 * The valid audio formats.
 */

#ifndef _SDL_audio_h
/* Audio format flags (defaults to LSB byte order) */
#  define SI_AUDIO_U8           0x0008 /* Unsigned 8-bit samples */
#  define SI_AUDIO_S8           0x8008 /* Signed 8-bit samples */
#  define SI_AUDIO_U16LSB       0x0010 /* Unsigned 16-bit samples */
#  define SI_AUDIO_S16LSB       0x8010 /* Signed 16-bit samples */
#  define SI_AUDIO_U16MSB       0x1010 /* As above, but big-endian byte order */
#  define SI_AUDIO_S16MSB       0x9010 /* As above, but big-endian byte order */
#  if SI_BYTE_ORDER == SI_LITTLE_ENDIAN
#    define SI_AUDIO_U16SYS     SI_AUDIO_U16LSB
#    define SI_AUDIO_S16SYS     SI_AUDIO_S16LSB
#  else
#    define SI_AUDIO_U16SYS     SI_AUDIO_U16MSB
#    define SI_AUDIO_S16SYS     SI_AUDIO_S16MSB
#  endif
#else
#  define SI_AUDIO_U8           AUDIO_U8
#  define SI_AUDIO_S8           AUDIO_S8
#  define SI_AUDIO_U16LSB       AUDIO_U16LSB
#  define SI_AUDIO_S16LSB       AUDIO_S16LSB
#  define SI_AUDIO_U16MSB       AUDIO_U16MSB
#  define SI_AUDIO_S16MSB       AUDIO_S16MSB
#  define SI_AUDIO_U16SYS       AUDIO_U16SYS
#  define SI_AUDIO_S16SYS       AUDIO_S16SYS
#endif /* _SDL_audio_h */


struct _SiAudioInfo
{
	Uint16  format;
	Uint32  rate;
	Uint8   channels;
	Uint8   reserved; /* padding */
};

struct _SiSoundDataInput
{
	int (* seek)  (void * context, int offset, int whence);
	int (* read)  (void * context, void * ptr, int size, int maxnum);
	int (* close) (void * context);
};

struct _SiSoundData
{
	SiLoader            * loader;       /* ptr to the loader */
	SiAudioInfo           requested;    /* the requested format */
	SiSoundDataFlags      flags;        /* flags */
	Uint8               * buffer;       /* the buffer with current data */
	Uint32                buffer_size;  /* the size of buffer */
	SiSoundDataInternal * internal;
};

extern SI_DECLSPEC
SiSoundData * si_sound_data_new (const SiSoundDataInput * input,
                                 void * input_context,
                                 const SiAudioInfo * requested,
                                 const char * extension, 
                                 const char * arg1, ...);

extern SI_DECLSPEC
SiSoundData * si_sound_data_new_arg (const SiSoundDataInput * input,
                                     void * input_context,
                                     const SiAudioInfo * requested,
                                     const char * extension,
                                     const SiLoaderArgument * args);

extern SI_DECLSPEC
SiSoundData * si_sound_data_new_stream (const SiSoundDataInput * input,
                                        void * input_context,
                                        const SiAudioInfo * requested,
                                        Uint32 buffer_size, 
                                        const char * extension, 
                                        const char * arg1, ...);

extern SI_DECLSPEC
SiSoundData * si_sound_data_new_stream_arg (const SiSoundDataInput * input,
                                            void * input_context,
                                            const SiAudioInfo * requested,
                                            Uint32 buffer_size, 
                                            const char * extension, 
                                            const SiLoaderArgument * args);

extern SI_DECLSPEC
void si_sound_data_free (SiSoundData * sound_data);

extern SI_DECLSPEC
Uint32 si_sound_data_decode (SiSoundData * sound_data);

/* Rewind the sound data. This is valid only for stremable
 * sound data. */
extern SI_DECLSPEC
int si_sound_data_rewind (SiSoundData * sound_data);

/*
 * Other stuff.
 */

extern SI_DECLSPEC
const char * si_get_error (void);

extern SI_DECLSPEC
int si_init ();

extern SI_DECLSPEC
void si_quit ();

#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* __si_h__ */

